#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jul  3 19:52:10 2017

@author: edanrein

v10-Oct-2018
-Cleaned up

v01-Aug-2018
-Cleaned up

v30-Aug-2017
-Notes on the transfer from MATLAB to python:
    1. No intersections became [] (empty) instead of NaN
    2. Still need to make the integrals for the non-uniform source case more accurate
    3. This file has all of the "basic" functions and is called by the other files.
"""

import numpy as np
import math as M
from pyppluss.fastqs import QuarticSolverVec as fastquartroots

def simpledist(u,v):
    """
    A simplified and more efficient 2-D euclidean distance calculation.
    The advantage over the built in (scipy) function is less checks.
    I do not check input validity.
    Assuming u,v are 2xn arrays, returns distance between n points.
    """
    return np.sqrt(np.sum((u-v)**2,0))

#%% Utility Functions: coordinate and argument swapping
def coord_swap (x_planet, y_planet, rotation_angle):
    """Rotation and translation of origin point
    
    Returns the new coordinates for the star's center after:
        1. Translation of the origin to (`x_planet`, `y_planet`)
        2. Rotation of the x axis `rotation_angle` radians
    
    Parameters
    ----------
    x_planet : ``1-D array``
        x coordinate of the planet's center
    y_planet : ``1-D array``
        y coordinate of the planet's center
    rotation_angle : ``float``
        Disk's rotation angle (in radians)

    Returns
    -------
    new_coords : ``2-D array``
        array of the (x,y) values of the new coordinates for the star's center.    
    """
    #This is because we rotate the axis rotation_angle, meaning the points go the other direcrtion
    #Using rotation matrix. Assuming rotation_angle is scalar, while the points may be vector.
    rotation_angle = -rotation_angle
    angle_cos = M.cos(rotation_angle)
    angle_sin = M.sin(rotation_angle)
    new_coords = np.dot(np.array([[angle_cos, -angle_sin],[angle_sin, angle_cos]]),-np.stack((x_planet,y_planet)))
    return new_coords

def get_coord_circle (arg,x_center,y_center,radius):
    """Calculates the argument for a point on a circle
    
    Calculates the coordinates for a point with argument `arg` on a circle with center (`x_center`, `y_center`) and radius `radius`.
    
    Parameters
    ----------
    arg : ``array``
        Argument of the point
    x_center : ``array``
        x coordinate of the circle's center
    y_center : ``array``
        y coordinate of the circle's center
    radius : ``array``
        radius of the circle
    
    Returns
    -------
    coords : ``array``
        a concatenated ``array`` (x,y) of the coordinates
    """
    x = x_center+radius*np.cos(arg)
    y = y_center+radius*np.sin(arg)
    return np.concatenate((x.reshape((1,-1)),y.reshape((1,-1))))

def get_coord_ellipse (arg,x_center,y_center,big_axis,small_axis):
    """Calculates the argument for a point on a ellipse
    
    Calculates the coordinates for a point with argument `arg` on a ellipse with center (`x_center`, `y_center`) and axes `big_axis` and `small_axis`.
    An ellipse behaves like a circle just not the same way (different r) in each axis.
    Parameters
    ----------
    arg : ``1-D array``
        Argument of the point (in radians)
    x_center : ``1-D array``
        x coordinate of the ellipse's center
    y_center : ``1-D array``
        y coordinate of the ellipse's center
    big_axis : ``1-D array``
        Big axis of the ellipse
    small_axis : ``1-D array``
        Small axis of the ellipse
    
    Returns
    -------
    coords : ``2-D array``
        a concatenated ``array`` (x,y) of the coordinates
    """
    k = big_axis/small_axis
    #Make sure atan is in the right quadrant
    angleF = np.arctan(k*np.tan(arg))
    #q4 and q1 are good with atan, only q2,q3 change quadrants inaccurately. This is a fix.
    angleT = angleF+(-np.sign(np.cos(arg))+1)*0.5*M.pi 
    x = x_center+big_axis*np.cos(angleT)
    y = y_center+small_axis*np.sin(angleT)
    return np.concatenate((x.reshape((1,-1)),y.reshape((1,-1))))
    
def get_arg (x, y, x_center, y_center):
    """Calculates the argument for a point relative to a center point
    
    Parameters
    ----------
    x : ``1-D array``
        x coordinate of the point
    y : ``1-D array``
        y coordinate of the point
    x_center : ``1-D array``
        x coordinate of the ellipse's center
    y_center : ``1-D array``
        y coordinate of the ellipse's center
    
    Returns
    -------
    coords : ``1-D array``
        The argument in radians, between 0 and 2*pi
    """
    arg = np.arctan2(y-y_center,x-x_center)
    #Make sure the angle is between 0 and 2*pi
    if np.isscalar(arg):
        if arg<0:
            arg += 2*M.pi
    else:
        arg[np.signbit(arg)] += 2*M.pi
    return arg

#%% Intersection Functions
def get_star_planet_intersection (x_star, y_star, radius_planet, tol):
    """Returns all the intersections between planet and star
    
    Returns all the intersections between:
    Star with center (`x_star`, `y_star`) and radius 1
    Planet with center (0,0) and radius `radius_planet`
    Using algorithm found on the internet - here: http://paulbourke.net/geometry/circlesphere/
    With the title **Intersection of two circles**
    Returns NaN in excess coordinates.
    
    Parameters
    ----------
    x_star : ``1-D array``
        x coordinate of the star's center
    y_star : ``1-D array``
        y coordinate of the star's center
    radius_planet : ``1-D array``
        Planet's radius relative to star's radius
    tol : ``float``
        Error tolerance parameter

    Returns
    -------
    intersections : ``2-D array``
        array of intersection points. Each point has its own column with the following data:
        x coordinate, y coordinate, star argument, planet argument
    """
    d = np.sqrt(x_star**2+y_star**2)
    #Checking existance of solutions
    star_planet_intersections = np.empty((len(x_star),4,2))
    star_planet_intersections[:] = np.NaN
    ind = np.logical_and(np.logical_and(d>abs(1-radius_planet), d<1+radius_planet), np.logical_not(d==0,radius_planet==1))
    if np.any(ind):
        d = d[ind]
        x_tmp = x_star[ind]
        y_tmp = y_star[ind]
        r_tmp = radius_planet[ind]
        #Calculating the solutions
        a = (1-r_tmp**2+d**2)/(2*d)
        h = np.sqrt(1-a**2)
        x2 = x_tmp-a*x_tmp/d
        y2 = y_tmp-a*y_tmp/d
        #Using np.concatenate to create a vector in the right shape.
        s1x = (x2-h*y_tmp/d).reshape(-1,1,1)
        s2x = (x2+h*y_tmp/d).reshape(-1,1,1)
        sx = np.concatenate((s1x,s2x),2)
        s1y = (y2+h*x_tmp/d).reshape(-1,1,1)
        s2y = (y2-h*x_tmp/d).reshape(-1,1,1)
        sy = np.concatenate((s1y,s2y),2)
        star_planet_intersections[ind,:2,:2] = np.concatenate((sx,sy),1)
        #Adding arguments
        mx = np.repeat(np.reshape(x_tmp,(len(x_tmp),1)),2,axis=1)
        my = np.repeat(np.reshape(y_tmp,(len(y_tmp),1)),2,axis=1)
        star_planet_intersections[ind,2,:2] = get_arg(star_planet_intersections[ind,0,:2], star_planet_intersections[ind,1,:2], mx, my)
        star_planet_intersections[ind,3,:2] = get_arg(star_planet_intersections[ind,0,:2], star_planet_intersections[ind,1,:2], 0, 0)
        #Checking duplicates
        veryclose = np.sum((star_planet_intersections[ind,:2,0]-star_planet_intersections[ind,:2,1])**2,1)<=tol
        star_planet_intersections[ind,:,1][veryclose] = np.NaN
    return star_planet_intersections

def get_star_disk_intersection (x_star, y_star, disk_radius, disk_inclination, tol):
    """Returns all the intersections between star and disk
    
    Returns all the intersections between:
    Star with center (`x_star`,`y_star`) and radius 1
    Disk with center (0,0), big axis `disk_radius` and small axis `disk-radius`*`disk_inclination`
    
    Parameters
    ----------
    x_star : ``1-D array``
        x coordinate of the star's center
    y_star : ``1-D array``
        y coordinate of the star's center
    disk_radius : ``1-D array``
        Disk's radius relative to star's radius
    disk_inclination : ``1-D array``
        Disk's inclination
    tol : ``float``
        Error tolerance parameter

    Returns
    -------
    intersections : ``2-D array``
        array of intersection points. Each point has its own column with the following data:
        x coordinate, y coordinate, star argument, disk argument
    """    
    #Calculate the coefficients of the polynomial
    #Polynomial in t using uniform trigonometric substitution (Weierstrass) for angle in disk.
    #NOTE: where a==0, it crashes.
    star_disk_intersections = np.empty([len(x_star),4,4])
    star_disk_intersections[:] = np.nan
    a = disk_radius
    b = a*np.cos(disk_inclination)
    num = x_star**2+y_star**2-1
    coeff_a = a**2+2*a*x_star+num
    coeff_b = -4*b*y_star
    coeff_c = -2*a**2+4*b**2+2*num
    coeff_d = coeff_b
    coeff_e = a**2-2*a*x_star+num
    specialcase = coeff_a==0
    specinds = np.nonzero(specialcase)[0]
    speclen = specinds.size

    if speclen>0:
        star_disk_intersections[specialcase,0,0]=-a[specialcase]
        star_disk_intersections[specialcase,1,0]=0.0
        for j in range(speclen):
            this_ind = specinds[j]
            t = np.empty(3)
            t[:] = np.nan
            tmp = np.roots([coeff_b[this_ind],coeff_c[this_ind],coeff_d[this_ind],coeff_e[this_ind]])
            t[:tmp.size] = tmp
            isnr = np.logical_not(t.imag==0)
            t[isnr] = np.nan
            #Convert from t to (x,y) coordinates
            x = a[this_ind]*(1-t**2)/(1+t**2)
            y = b[this_ind]*2*t/(1+t**2)
            star_disk_intersections[this_ind,0,1:] = x.real.T
            star_disk_intersections[this_ind,1,1:] = y.real.T
            
    #Every row of t corresponds to an input.
    regcase = np.logical_not(specialcase)
    t = fastquartroots(coeff_a[regcase].astype(complex),coeff_b[regcase].astype(complex),coeff_c[regcase].astype(complex),coeff_d[regcase].astype(complex),coeff_e[regcase].astype(complex))
    isnr = np.logical_not(t.imag==0)
    t[isnr] = np.nan
    #Convert from t to (x,y) coordinates
    x = a[regcase]*(1-t**2)/(1+t**2)
    y = b[regcase]*2*t/(1+t**2)

    star_disk_intersections[regcase,0,:] = x.real.T
    star_disk_intersections[regcase,1,:] = y.real.T
    cond = specialcase
    cond[regcase] = np.any(np.logical_not(isnr),0)
    #Calculating arguments
    mx = np.repeat(np.reshape(x_star[cond],(len(x_star[cond]),1)),4,axis=1)
    my = np.repeat(np.reshape(y_star[cond],(len(y_star[cond]),1)),4,axis=1)
    star_disk_intersections[cond,2,:] = get_arg(star_disk_intersections[cond,0,:],star_disk_intersections[cond,1,:],mx,my); #Star argument
    star_disk_intersections[cond,3,:] = get_arg(star_disk_intersections[cond,0,:],star_disk_intersections[cond,1,:],0,0); #Disk argument
    return star_disk_intersections

def get_disk_planet_intersection (radius_planet, disk_radius, disk_inclination, tol):
    """Returns all the intersections between planet and disk
    
    Returns all the intersections between:
    Planet with center (0,0) and radius `radius_planet`
    Disk with center (0,0), big axis `disk_radius` and small axis `disk-radius` * cos(`disk_inclination`)
    Using equation from Wolfram.
    
    Parameters
    ----------
    radius_planet : ``1-D array``
        Planet's radius relative to star's radius
    disk_radius : ``1-D array``
        Disk's radius relative to star's radius
    disk_inclination : ``float``
        Disk's inclination (in radians)
    tol : ``float``
        Error tolerance parameter

    Returns
    -------
    intersections : ``2-D array``
        array of intersection points. Each point has its own column with the following data:
        x coordinate, y coordinate, planet/disk argument
    """    
    n = len(radius_planet)
    disk_planet_intersections = np.zeros([n,3,4])
    b = disk_radius*M.cos(disk_inclination)
    cond = np.logical_or(b>radius_planet,disk_radius<radius_planet)
    disk_planet_intersections[cond,:,:] = np.nan
    cond = np.logical_not(cond)
    n = len(radius_planet[cond])
    rp = radius_planet[cond]
    rd = disk_radius[cond]
    b = b[cond]
    #Calculations
    disk_planet_intersections[cond,0,:2] = np.repeat(rd*np.sqrt((rp-b)*(rp+b)/(rd**2-b**2)),2).reshape((n,2))
    disk_planet_intersections[cond,0,2:4] = -disk_planet_intersections[cond,0,:2]
    disk_planet_intersections[cond,1,:3:2] = np.repeat(b*np.sqrt((rd**2-rp**2)/(rd**2-b**2)),2).reshape((n,2))
    disk_planet_intersections[cond,1,1:4:2] = -disk_planet_intersections[cond,1,:3:2]
    #Void duplicates when x value is too small
    disk_planet_intersections[cond,:,2:][2*disk_planet_intersections[cond,0,0]<=tol,:,:] = np.nan
    #Void duplicates when y value is too small
    disk_planet_intersections[cond,:,:3:2][2*disk_planet_intersections[cond,1,0]<=tol,:,:] = np.nan
    #Planet/Disk argument
    disk_planet_intersections[cond,2,:] = get_arg(disk_planet_intersections[cond,0,:],disk_planet_intersections[cond,1,:],0,0)
    return disk_planet_intersections


#%% Sector area calculation helpers
def triangle_area (x1,y1,x2,y2,x3,y3):
    """Calculates triangle area
    
    The calculation is made based on the (x,y) positions of the vertices.
    
    Parameters
    ----------
    x1 : ``1-D array``
        x coordinate of the first vertex
    y1 : ``1-D array``
        y coordinate of the first vertex
    x2 : ``1-D array``
        x coordinate of the second vertex
    y2 : ``1-D array``
        y coordinate of the second vertex
    x3 : ``1-D array``
        x coordinate of the third vertex
    y3 : ``1-D array``
        y coordinate of the third vertex
        
    Returns
    -------
    area : ``1-D array``
        Area of triangle defined by the three veritces
    """
    #This is a simple case of the polygon area equation.
    area = 0.5*abs(x1*(y3-y2)+x2*(y1-y3)+x3*(y2-y1))
    return area

def ellipse_to_x (big_axis,small_axis,angle):
    """Calculates sector area of the ellipse
    
    This is a helper function intended to aid sector_area_disk in calculations
    assuming all angles are between 0 and 2pi
    
    Parameters
    ----------
    big_axis : ``1-D array``
        The long axis of the ellipse
    small_axis : ``1-D array``
        The short axis of the ellipse
    angle : ``1-D array``
        Angle of the point relative to the positive direction of the long axis
    
    Returns
    -------
    area : ``1-D array``
        Area of sector between the positive direction of the long axis and angle
    """
    #The Formula is present in the paper
    rangle = np.round(angle/np.pi)*np.pi
    area = big_axis*small_axis*rangle*0.5+0.5*big_axis*small_axis*np.arctan(big_axis/small_axis*np.tan(angle-rangle))
    return area
    
#%% Sector area calculations
def sector_area_planet (arg1, arg2, radius_planet):
    """Calculates sector area of the planet
    
    The sector area is between `arg1` and `arg2` in circle with radius `radius planet` and center at origin.
    The calculation is done as required for the sector based algorithm.
    We assume that the area is in the **counterclockwise** direction from `arg1` to `arg2`
    
    Parameters
    ----------
    arg1 : ``1-D array``
        The angle to start going counterclockwise from (in radians)
    arg2 : ``1-D array``
        The angle end of the sector (in radians)
    radius_planet : ``1-D array``
        The radius of the planet
    Returns
    -------
    area : ``1-D array``
        The requested sector area
    """
    if np.isscalar(arg1):        
        if arg2<arg1 :
            area = M.pi*radius_planet**2-0.5*radius_planet**2*(arg1-arg2)
        else:
            area = 0.5*radius_planet**2*(arg2-arg1)
    else:
        area = 0.5*radius_planet**2*(arg2-arg1)
        cond = arg2<arg1
        area[cond] = M.pi*radius_planet[cond]**2+area[cond]
    return area

def sector_area_star (arg1, arg2):
    """Calculates sector area of the star
    
    The sector area is between `arg1` and `arg2` in circle with radius 1 and center at (`x_star`, `y_star`).
    The calculation is done as required for the sector based algorithm.
    We assume that the area is in the **counterclockwise** direction from `arg1` to `arg2`
    
    Parameters
    ----------
    arg1 : ``1-D array``
        The angle to start going counterclockwise from (in radians)
    arg2 : ``1-D array``
        The angle end of the sector (in radians)
    x_star : ``1-D array``
        x coordinate of the star's center
    y_star : ``1-D array``
        y coordinate of the star's center
    Returns
    -------
    area : ``1-D array``
        The requested sector area
    """
    secArea = 0.5*(arg2-arg1)
    cond = arg2<arg1
    secArea[cond] = M.pi+secArea[cond]
    return secArea

def sector_area_disk (arg1, arg2, disk_radius, disk_inclination):
    """Calculates sector area of the disk
    
    The sector area is between `arg1` and `arg2` in ellipse with radius `disk_radius`, inclination `disk_inclination` and center at origin.
    The calculation is done as required for the sector based algorithm.
    We assume that the area is in the **counterclockwise** direction from `arg1` to `arg2`
    
    Parameters
    ----------
    arg1 : ``1-D array``
        The angle to start going counterclockwise from (in radians), between 0 and 2pi
    arg2 : ``1-D array``
        The angle end of the sector (in radians), between 0 and 2pi
    disk_radius : ``1-D array``
        The disk's big axis
    disk_inclination : ``float``
        the inclination of the disk (in radians)
    Returns
    -------
    area : ``1-D array``
        The requested sector area
    """
    if np.isscalar(arg1):
        if arg2>arg1:
            area = ellipse_to_x(disk_radius,disk_radius*M.cos(disk_inclination),arg2)-ellipse_to_x(disk_radius,disk_radius*M.cos(disk_inclination),arg1)
        else:
            area = M.pi*disk_radius**2*M.cos(disk_inclination)-(ellipse_to_x(disk_radius,disk_radius*M.cos(disk_inclination),arg1)-ellipse_to_x(disk_radius,disk_radius*M.cos(disk_inclination),arg2))
    else:
        area = ellipse_to_x(disk_radius,disk_radius*M.cos(disk_inclination),arg2)-ellipse_to_x(disk_radius,disk_radius*M.cos(disk_inclination),arg1)
        cond = arg2<arg1
        area[cond] = M.pi*disk_radius[cond]**2*M.cos(disk_inclination)+area[cond]
    return area

#%% Border helper
def get_dha_border (radius_planet, disk_radius, disk_inclination, x_star, y_star, star_planet_intersections, star_disk_intersections, disk_planet_intersections, tol):
    """Finds which intersection points are on the dha border
    
    Returns the dha_border as required by double hidden area calculations
    
    Parameters
    ----------
    radius_planet : ``1-D array``
        Planet's radius
    disk_radius : ``1-D array``
        Disk's radius
    disk_inclination : ``float``
        Disk's inclination (in radians)
    x_star : ``1-D array``
        x coordinate of the star's center
    y_star : ``1-D array``
        y coordinate of the star's center
    star_planet_intersections : ``2-D array``
        the array of star-planet intersections as returned by get_star_planet_intersection
    star_disk_intersections : ``2-D array``
        the array of star-disk intersections as returned by get_star_disk_intersection
    disk_planet_intersections : ``2-D array``
        the array of disk-planet intersections as returned by get_disk_planet_intersection
    tol : ``float``, optional
        Error tolerance parameter
    
    Returns
    -------
    dha_border : ``2-D array``
        Points that are on the border of the dha. each point is a column consisting of x,y coordinates, argument relative to planet's center and curves.
    """
    # 0=star, 1=planet, 2=disk
    #These identifiers are on row  #3 and #4
    #Taking only the argument relative to the planet's center
    #Stacking them all together on the right dimension
    n = star_planet_intersections.shape[0]
    this_spi = np.concatenate((star_planet_intersections[:,[0,1,3],:],np.zeros((n,1,2)),np.ones((n,1,2)),np.nan*np.empty((n,1,2))),1)
    this_sdi = np.concatenate((star_disk_intersections[:,[0,1,3],:],np.zeros((n,1,4)),2*np.ones((n,1,4)),np.nan*np.empty((n,1,4))),1)
    this_dpi = np.concatenate((disk_planet_intersections,np.ones((n,1,4)),2*np.ones((n,1,4)),np.nan*np.empty((n,1,4))),1)
    all_intersections = np.concatenate((this_spi,this_sdi,this_dpi),-1)
    if all_intersections.ndim<3:
        all_intersections = np.reshape(all_intersections, (1,all_intersections.shape[0], all_intersections.shape[1]))
    #Look for dha border only when there are intersections
    iexist = np.logical_not(np.all(np.isnan(all_intersections[:,0,:]),-1))
    test_intersections = all_intersections[iexist,:,:]
    test_intersections.shape = (-1,all_intersections.shape[1], all_intersections.shape[2])
    if test_intersections.size>0:
        #There are 10 intersection points (including nans)
        sdist_edge = np.sqrt((test_intersections[:,0,:]-np.repeat(x_star[iexist].reshape((-1,1)),10,1))**2+(test_intersections[:,1,:]-np.repeat(y_star[iexist].reshape((-1,1)),10,1))**2)-1
        pdist_edge = np.sqrt(test_intersections[:,0,:]**2+test_intersections[:,1,:]**2)-np.repeat(radius_planet[iexist].reshape((-1,1)),10,1)
        #Signbit is the same as <0
        is_in_star = np.signbit(sdist_edge-tol)
        is_in_planet = np.signbit(pdist_edge-tol)
        disk_dist_1 = np.sqrt((test_intersections[:,0,:]-np.repeat(disk_radius[iexist].reshape((-1,1)),10,1)*M.sin(disk_inclination))**2+test_intersections[:,1,:]**2)
        disk_dist_2 = np.sqrt((test_intersections[:,0,:]+np.repeat(disk_radius[iexist].reshape((-1,1)),10,1)*M.sin(disk_inclination))**2+test_intersections[:,1,:]**2)
        ddist_edge = disk_dist_1+disk_dist_2-2*np.repeat(disk_radius[iexist].reshape((-1,1)),10,1)
        is_in_disk = np.signbit(ddist_edge-tol)
        is_dha_border = np.logical_and(np.logical_and(is_in_star,is_in_planet),is_in_disk)
        #Numpy>=1.10.0, so swapaxes is view and not copy
        np.swapaxes(test_intersections,0,1)[:,np.logical_not(is_dha_border)] = np.nan
        dexist = np.logical_not(np.all(np.isnan(test_intersections[:,0,:]),-1))
        #dborder and test_intersections are not seperate arrays from all_intersections.
        #They seem to function as pointers?
        dborder = test_intersections[dexist,:,:]
        if dborder.size>0:
            #Adding third curve if very close
            dtriple = np.signbit(abs(ddist_edge[dexist,:2])-tol)
            ptriple = np.signbit(abs(pdist_edge[dexist,2:6])-tol)
            striple = np.signbit(abs(sdist_edge[dexist,6:])-tol)
            np.swapaxes(dborder[:,3:,:2],1,2)[dtriple,:] = np.array([0,1,2])
            np.swapaxes(dborder[:,3:,2:6],1,2)[ptriple,:] = np.array([0,1,2])
            np.swapaxes(dborder[:,3:,6:],1,2)[striple,:] = np.array([0,1,2])
        #Ensuring no pointer issues (again)
        test_intersections[dexist,:,:] = dborder
        all_intersections[iexist,:,:] = test_intersections
    else:
        all_intersections[:] = np.nan
    return all_intersections



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Sep 29 12:27:45 2017

@author: edanrein
Vectorized FastQS as described in PeterStrobach (2010), Journal of Computational and Applied Mathematics 234
Based on MATLAB port by Aviv Ofir

v10-Oct-2018
-Cleaned up

v03-Sep-2018
-Increased MaxIter back to 50, to handle special cases.

v02-Aug-2018
-Cleaned up
-Reduced MaxIter back to 16
"""
import numpy as np

def QuarticSolverVec(a,b,c,d,e):
    """
     function [x1, x2, x3, x4]=QuarticSolverVec(a,b,c,d,e)
     v.0.2 - Python Port
     - Added condition in size sorting to avoid floating point errors.
     - Removed early loop abortion when stuck in loop (Inefficient)
     - Improved numerical stability of analytical solution
     - Added code for the case of S==0
     ============================================
     v.0.1 - Nearly identical to QuarticSolver v. 0.4, the first successful vectorized implimentation 
             Changed logic of ChosenSet to accomudate simultaneous convergence of sets 1 & 2
           - Note the periodicity in nearly-convergent solutions can other
             than four (related to text on step 4 after table 3). examples:
             period of 5: [a,b,c,d,e]=[0.111964240308252 -0.88497524334712 -0.197876116344933 -1.07336408259262 -0.373248675102065];
             period of 6: [a,b,c,d,e]=[-1.380904438798326 0.904866918945240 -0.280749330818231 0.990034312758900 1.413106456228119];
             period of 22: [a,b,c,d,e]=[0.903755513939902 0.490545114637739 -1.389679906455410 -0.875910689438623 -0.290630547104907];
             Therefore condition was changed from epsilon1(iiter)==0 to epsilon1(iiter)<8*eps (and similarl for epsilon2)
           - Special case criterion of the analytical formula was changed to
             ind=abs(4*Delta0**3./Delta1**2)<2*eps;  (instead of exact zero)
           - vectorized
     ============================================
     - Solves for the x1-x4 roots of the quartic equation y(x)=ax^4+bx^3+cx^2+dx+e.
       Multiple eqations can be soved simultaneously by entering same-sized column vectors on all inputs.
     - Note the code immediatly tanslates the input parameters ["a","b","c","d","e"] to the reference paper parameters [1,a,b,c,d] for consistency,
       and the code probably performes best when "a"=1.
    
    Parameters
    ----------
    a,b,c,d,e : ``1-D arrays``
        Quartic polynomial coefficients
    
    Returns
    ------
    - x1-x4 : ``2-D array``
        Concatenated array of the polynomial roots. The function always returns four (possibly complex) values. Multiple roots, if exist, are given multiple times. An error will result in four NaN values.
        No convergence may result in four inf values (still?)
    
    Reference: 
    Peter Strobach (2010), Journal of Computational and Applied Mathematics 234
        http://www.sciencedirect.com/science/article/pii/S0377042710002128
    """
#    MaxIter=16;
    MaxIter=50;
    eps = np.finfo(float).eps
    #INPUT CONTROL
    #Note: not all input control is implemented.
    # all-column vectors only
#    if size(a,1)~=size(b,1) or size(a,1)~=size(c,1) or size(a,1)~=size(d,1) or size(a,1)~=size(e,1) or ...
#       size(a,2)~=1 or size(b,2)~=1 or size(c,2)~=1 or size(d,2)~=1 or size(e,2)~=1:
#        fprintf('ERROR: illegal input parameter sizes.\n');
#        x1=inf; x2=inf; x3=inf; x4=inf;    
#        return
    
    # translate input variables to the paper's
    if np.any(a==0):
       print('ERROR: a==0. Not a quartic equation.\n')
       x1=np.NaN; x2=np.NaN; x3=np.NaN; x4=np.NaN;    
       return x1,x2,x3,x4
    else:
        input_a=a;
        input_b=b;
        input_c=c;
        input_d=d;
        input_e=e;
        a=input_b/input_a;
        b=input_c/input_a;
        c=input_d/input_a;
        d=input_e/input_a;
    
    # PRE-ALLOCATE MEMORY
    # ChosenSet is used to track which input set already has a solution (=non-zero value)
    ChosenSet=np.zeros_like(a);
    x1 = np.empty_like(a,complex)
    x1[:] = np.nan
    x2=x1.copy(); x3=x1.copy(); x4=x1.copy(); x11=x1.copy(); x12=x1.copy(); x21=x1.copy(); x22=x1.copy(); alpha01=x1.copy(); alpha02=x1.copy(); beta01=x1.copy(); beta02=x1.copy(); gamma01=x1.copy(); gamma02=x1.copy(); delta01=x1.copy(); delta02=x1.copy(); e11=x1.copy(); e12=x1.copy(); e13=x1.copy(); e14=x1.copy(); e21=x1.copy(); e22=x1.copy(); e23=x1.copy(); e24=x1.copy(); alpha1=x1.copy(); alpha2=x1.copy(); beta1=x1.copy(); beta2=x1.copy(); gamma1=x1.copy(); gamma2=x1.copy(); delta1=x1.copy(); delta2=x1.copy(); alpha=x1.copy(); beta=x1.copy(); gamma=x1.copy(); delta=x1.copy();
    # check multiple roots -cases 2 & 3. indexed by ChosenSet=-2
    test_alpha=0.5*a;
    test_beta=0.5*(b-test_alpha**2);
    test_epsilon=np.stack((c-2*test_alpha*test_beta, d-test_beta**2)).T;
    ind=np.all(test_epsilon==0,1);
    if np.any(ind):
        x1[ind], x2[ind]=SolveQuadratic(np.ones_like(test_alpha[ind]),test_alpha[ind],test_beta[ind]);
        x3[ind]=x1[ind]; x4[ind]=x2[ind];
        ChosenSet[ind]=-2;
    
    # check multiple roots -case 4. indexed by ChosenSet=-4
    i=ChosenSet==0;
    x11[i], x12[i]=SolveQuadratic(np.ones(np.sum(i)),a[i]/2,b[i]/6);
    x21[i]=-a[i]-3*x11[i];    
    test_epsilon[i,:2]=np.stack((c[i]+x11[i]**2*(x11[i]+3*x21[i]), d[i]-x11[i]**3*x21[i])).T;
    ind[i]=np.all(test_epsilon[i]==0,1);
    if np.any(ind[i]):
        x1[ind[i]]=x11[ind[i]]; x2[ind[i]]=x11[ind[i]]; x3[ind[i]]=x11[ind[i]]; x4[ind[i]]=x12[ind[i]];
        ChosenSet[ind[i]]=-4;
    x22[i]=-a[i]-3*x12[i];
    test_epsilon[i,:2]=np.stack((c[i]+x12[i]**2*(x12[i]+3*x22[i]), d[i]-x12[i]**3*x22[i])).T;
    ind[i]=np.all(test_epsilon[i]==0,1);
    if np.any(ind[i]):
        x1[ind[i]]=x21[ind[i]]; x2[ind[i]]=x21[ind[i]]; x3[ind[i]]=x21[ind[i]]; x4[ind[i]]=x22[ind[i]];
        ChosenSet[ind[i]]=-4;
    # General solution
    # initilize
    epsilon1=np.empty((np.size(a),MaxIter))
    epsilon1[:]=np.inf
    epsilon2=epsilon1.copy();
    
    i=ChosenSet==0;
    fi=np.nonzero(i)[0];
    x=np.empty((fi.size,4),complex)
    ii = np.arange(fi.size)
    #Calculate analytical root values
    x[:,0], x[:,1], x[:,2], x[:,3]=AnalyticalSolution(np.ones(np.sum(i)),a[i],b[i],c[i],d[i],eps);
    #Sort the roots in order of their size
    ind=np.argsort(abs(x))[:,::-1]; #'descend'
    x1[i]=x.flatten()[4*ii+ind[:,0]];
    x2[i]=x.flatten()[4*ii+ind[:,1]];
    x3[i]=x.flatten()[4*ii+ind[:,2]];
    x4[i]=x.flatten()[4*ii+ind[:,3]];
    #Avoiding floating point errors.
    #The value chosen is somewhat arbitrary. See Appendix C for details.
    ind = abs(x1)-abs(x4)<8*10**-12;
    x2[ind] = np.conj(x1[ind])
    x3[ind] = -x1[ind]
    x4[ind] = -x2[ind]
    #Initializing parameter values
    alpha01[i]=-np.real(x1[i]+x2[i]);
    beta01[i]=np.real(x1[i]*x2[i]);
    alpha02[i]=-np.real(x2[i]+x3[i]);
    beta02[i]=np.real(x2[i]*x3[i]);
    gamma01[i], delta01[i]=FastGammaDelta(alpha01[i],beta01[i],a[i],b[i],c[i],d[i]);
    gamma02[i], delta02[i]=FastGammaDelta(alpha02[i],beta02[i],a[i],b[i],c[i],d[i]);
    
    alpha1[i]=alpha01[i]; beta1[i]=beta01[i]; gamma1[i]=gamma01[i]; delta1[i]=delta01[i];
    alpha2[i]=alpha02[i]; beta2[i]=beta02[i]; gamma2[i]=gamma02[i]; delta2[i]=delta02[i];
    
    #Backward Optimizer Outer Loop
    e11[i]=a[i]-alpha1[i]-gamma1[i];
    e12[i]=b[i]-beta1[i]-alpha1[i]*gamma1[i]-delta1[i];
    e13[i]=c[i]-beta1[i]*gamma1[i]-alpha1[i]*delta1[i];
    e14[i]=d[i]-beta1[i]*delta1[i];
    
    e21[i]=a[i]-alpha2[i]-gamma2[i];
    e22[i]=b[i]-beta2[i]-alpha2[i]*gamma2[i]-delta2[i];
    e23[i]=c[i]-beta2[i]*gamma2[i]-alpha2[i]*delta2[i];
    e24[i]=d[i]-beta2[i]*delta2[i];
    iiter=0;
    while iiter<MaxIter and np.any(ChosenSet[i]==0):
        i=np.nonzero(ChosenSet==0)[0];
        
        alpha1[i], beta1[i], gamma1[i], delta1[i], e11[i], e12[i], e13[i], e14[i], epsilon1[i,iiter]=BackwardOptimizer_InnerLoop(a[i],b[i],c[i],d[i],alpha1[i],beta1[i],gamma1[i],delta1[i],e11[i],e12[i],e13[i],e14[i]);
        alpha2[i], beta2[i], gamma2[i], delta2[i], e21[i], e22[i], e23[i], e24[i], epsilon2[i,iiter]=BackwardOptimizer_InnerLoop(a[i],b[i],c[i],d[i],alpha2[i],beta2[i],gamma2[i],delta2[i],e21[i],e22[i],e23[i],e24[i]);
    
        j = np.ones_like(a[i])
        j[(epsilon2[i,iiter]<epsilon1[i,iiter]).flatten()] = 2
        BestEps = np.nanmin(np.stack([epsilon1[i,iiter].flatten(), epsilon2[i,iiter].flatten()]),0);
        ind=BestEps<8*eps;
        ChosenSet[i[ind]]=j[ind];
        ind=np.logical_not(ind);
#        if iiter>0 and np.any(ind):
#            ii=i[ind];
#            LimitCycleReached = np.empty((ii.size,2),bool)
#            LimitCycleReached[:,0] = np.any(epsilon1[ii,:iiter]==epsilon1[ii,iiter],0)
#            LimitCycleReached[:,1] = np.any(epsilon2[ii,:iiter]==epsilon2[ii,iiter],0)
##            LimitCycleReached=[any(bsxfun(@eq,epsilon1(i(ind),max(1,iiter-4):max(1,iiter-1)),epsilon1(i(ind),iiter)),2) any(bsxfun(@eq,epsilon2(i(ind),max(1,iiter-4):max(1,iiter-1)),epsilon2(i(ind),iiter)),2)];
#            ChosenSet[ii[np.logical_and(LimitCycleReached[:,0] , np.logical_not(LimitCycleReached[:,1]))]]=1;
#            ChosenSet[ii[np.logical_and(LimitCycleReached[:,1] , np.logical_not(LimitCycleReached[:,0]))]]=2;
##            ChosenSet(ii(~LimitCycleReached(:,1) & LimitCycleReached(:,2)))=2;
##            ind=find(ind);
#            cond = np.logical_and(LimitCycleReached[:,1],LimitCycleReached[:,0])
#            ChosenSet[ii[cond]]=j[ind][cond]
##            ChosenSet(ii(LimitCycleReached(:,1) & LimitCycleReached(:,2)))=j(ind(LimitCycleReached(:,1) & LimitCycleReached(:,2)));
        iiter=iiter+1;
        
    #Checking which of the chains is relevant
    i=np.nonzero(ChosenSet==0)[0];
    ind=epsilon1[i,-1]<epsilon2[i,-1];
#    ind=np.logical_and(epsilon1[i,-1]<epsilon2[i,-1],np.logical_not(np.isnan(epsilon2[i,-1])));
    ChosenSet[i[ind]]=1;
    ChosenSet[i[np.logical_not(ind)]]=2;
    
    # Output
    i=ChosenSet==1;
    alpha[i]=alpha1[i];
    beta[i]=beta1[i];
    gamma[i]=gamma1[i];
    delta[i]=delta1[i];
    
    i=ChosenSet==2;
    alpha[i]=alpha2[i];
    beta[i]=beta2[i];
    gamma[i]=gamma2[i];
    delta[i]=delta2[i];
    
    i=ChosenSet>0;
    x1[i], x2[i]=SolveQuadratic(np.ones(np.sum(i)),alpha[i],beta[i]);
    x3[i], x4[i]=SolveQuadratic(np.ones(np.sum(i)),gamma[i],delta[i]);

    return np.array([x1,x2,x3,x4])


def AnalyticalSolution(a,b,c,d,e,eps):
    """
    Source: https://en.wikipedia.org/wiki/Quartic_function#General_formula_for_roots
    Calculates the value of the analytical solution.
    """
    p=(8*a*c-3*b**2)/(8*a**2)
    q=(b**3-4*a*b*c+8*a**2*d)/(8*a**3)
    
    Delta0=c**2-3*b*d+12*a*e
    Delta1=2*c**3 -9*b*c*d +27*b**2*e +27*a*d**2 -72*a*c*e
    Q = np.empty_like(Delta1, complex)
    cond = Delta1>=0
    Q[cond]=(0.5*(Delta1[cond]+np.sqrt(Delta1[cond]**2-4*Delta0[cond]**3)))**(1/3)
    #Improved numerical stability of calculation of Q by noting (when d1<0):
    #(a+b)(a-b)=a**2-b**2; here (d1+sqrt(...))(d1-sqrt(...))=4*d0**3.
    cond = Delta1<0
    Q[cond]=Delta0[cond]*(2/(Delta1[cond]-np.sqrt(Delta1[cond]**2-4*Delta0[cond]**3)))**(1/3)
    #In this case we need to set the sign of sqrt equal to that of delta1
    ind = abs(Delta0)<eps
    Q[ind] = Delta1[ind]**(1/3)
    #The case of triple root+singular is not properly handled by the code.
#    cond = np.logical_and(ind,Delta1<eps)
#    aa = a[cond]; bb=b[cond]; cc=c[cond];dd = d[cond];
#    xparam = dd-bb*cc/(12*aa)-0.5*bb/aa*(5/6*cc-bb**2/(4*aa))
#    x1[cond] = xparam/(cc/(6*aa)*(5/6*cc-bb**2/(4*aa))-e[cond])
#    x2[cond] = x1[cond]
#    x3[cond] = x1[cond]
#    x4[cond] = -bb/aa-3*x1[cond]
#    cond = np.logical_not(cond)
#    ind=abs(4*Delta0**3)<2*eps*abs(Delta1**2)    
#    if np.any(ind):
#        ind2=np.zeros_like(a,bool);
#        Delta = np.empty_like(Q, complex)
#        Delta[ind]=256*a[ind]**3*e[ind]**3-192*a[ind]**2*b[ind]*d[ind]*e[ind]**2-128*a[ind]**2*c[ind]**2*e[ind]**2+144*a[ind]**2*c[ind]*d[ind]**2*e[ind]-27*a[ind]**2*d[ind]**4
#        +144*a[ind]*b[ind]**2*c[ind]*e[ind]**2-6*a[ind]*b[ind]**2*d[ind]**2*e[ind]-80*a[ind]*b[ind]*c[ind]**2*d[ind]*e[ind]+18*a[ind]*b[ind]*c[ind]*d[ind]**3+16*a[ind]*c[ind]**4*e[ind] 
#        -4*a[ind]*c[ind]**3*d[ind]**2-27*b[ind]**4*e[ind]**2+18*b[ind]**3*c[ind]*d[ind]*e[ind]-4*b[ind]**3*d[ind]**3-4*b[ind]**2*c[ind]**3*e[ind]+b[ind]**2*c[ind]**2*d[ind]**2;
#        ind2[ind]=np.logical_not(Delta[ind]==0);
#        if np.any(Delta[np.logical_and(ind,ind2)]):
#            Q[np.logical_and(ind, ind2)]=(Delta1[np.logical_and(ind, ind2)])**(1/3);
    
    S=0.5*np.sqrt(-2/3*p+1/(3*a)*(Q+Delta0/Q))
    cond = S==0
    Q[cond] *= np.exp(1j*np.pi/3)
    S=0.5*np.sqrt(-2/3*p+1/(3*a)*(Q+Delta0/Q))
    cond = S==0
    Q[cond] *= np.exp(1j*np.pi/3)
    S=0.5*np.sqrt(-2/3*p+1/(3*a)*(Q+Delta0/Q))
    cond = S==0
    if np.any(cond):
        raise NotImplementedError("A root is (probably) singular or repeats 3 times. Code does not treat it (yet)")
    x1=-b/(4*a)-S+0.5*np.sqrt(-4*S**2-2*p+q/S);
    x2=-b/(4*a)-S-0.5*np.sqrt(-4*S**2-2*p+q/S);
    x3=-b/(4*a)+S+0.5*np.sqrt(-4*S**2-2*p-q/S);
    x4=-b/(4*a)+S-0.5*np.sqrt(-4*S**2-2*p-q/S);
    return x1,x2,x3,x4
   
def FastGammaDelta(alpha0,beta0,a,b,c,d):
    # Table 3
    phi1=1+alpha0**2+beta0**2;
    phi2=alpha0*(1+beta0);
    c1=a-alpha0+alpha0*(b-beta0)+beta0*c;
    c2=b-beta0+alpha0*c+beta0*d;
    L1=np.sqrt(phi1);
    L3=phi2/L1;
    L2=np.sqrt(phi1-phi2**2/phi1);
    y1=c1/L1;
    y2=(c2-y1*L3)/L2;
    delta0=y2/L2;
    gamma0=(y1-delta0*L3)/L1;
    return gamma0, delta0

def BackwardOptimizer_InnerLoop(a,b,c,d,alpha,beta,gamma,delta,e1,e2,e3,e4):
    U23=alpha-gamma;
    U33=beta-delta-gamma*U23;
    L43=-delta*U23/U33;
    U44=beta-delta-L43*U23;
   
    x1=e1;
    x2=e2-gamma*x1;
    x3=e3-delta*x1-gamma*x2;
    x4=e4-delta*x2-L43*x3;
    y4=x4/U44;
    y3=(x3-U23*y4)/U33;
    y2=x2-U23*y3-y4;
    y1=x1-y3;
       
    alpha=alpha+y1;
    beta=beta+y2;
    gamma=gamma+y3;
    delta=delta+y4;
   
    e1=a-alpha-gamma;
    e2=b-beta-alpha*gamma-delta;
    e3=c-beta*gamma-alpha*delta;
    e4=d-beta*delta;   
    epsilon=abs(e1)+abs(e2)+abs(e3)+abs(e4);
    return alpha, beta, gamma, delta, e1, e2, e3, e4, epsilon

def SolveQuadratic(a,b,c):
    # Chapter 5.6 from Numerical Recepies
    i=np.all(np.imag(np.stack([a, b, c]))==0,0);
    ii = b==0
    q = np.empty_like(a,complex)
    q[i]=-0.5*(b[i]+np.sign(b[i])*np.sqrt(b[i]**2-4*a[i]*c[i]));
    #Correct for sign(b)=0
    q[ii]=-0.5*np.sqrt(-4*a[ii]*c[ii])
    i=np.logical_not(i);
    if np.any(i):
        s = np.empty_like(a)
        s[i]=abs(np.real(np.conj(b[i])*np.sqrt(b[i]**2-4*a[i]*c[i])));
#        s[i[s[i]<0]]=-s[i[s[i]<0]];
        q[i]=-0.5*(b[i]+s[i]);
    
    i = q==0
    if np.any(i):
        x1=np.zeros_like(a,complex);x2=np.zeros_like(a,complex);
        i = np.logical_not(i)
        x1[i]=q[i]/a[i];
        x2[i]=c[i]/q[i];
    else:
        x1=q/a;
        x2=c/q;
    return x1,x2


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Aug  8 18:12:03 2017

@author: edanrein
This file has finalized models and a version ready for celerite and data generation.

v10-Oct-2018
-Cleaned Up
-Note: some functions are only useful when using commented-out debugging-related parts of the code.

v02-Aug-2018
-Cleaned up

v25-Jun-2018
-Modified celerite model to reflect Aviv's recommendations (re:normalization)
-Note: To avoid bugs, did not "clean up" past variables probably no longer used but you never know

v4-Jun-2018
-Fixed: for mucrit!=0, algorithm was incorrect in using uniform source
-Added: Support for two different Gaussian Quadrature orders: one for ingress/egress and the other for the rest.

v26-May-2018
-Implemented sub-intervals splitting
-Fixed: Bug when y_star=0 in Quartic eq. for min-max disk-star distance
-Fixed: Bug when loss of accuracy in conversion form r->t caused issues with subinterval calculations

"""
import numpy as np
import math as M
from pyppluss.base_functions import coord_swap, get_disk_planet_intersection, get_star_disk_intersection, get_star_planet_intersection
from pyppluss.polygon_plus_segments import intersection_area as double_hidden_area
from pyppluss.polygon_plus_segments import handler, Star, Planet, Disk, border_sort
from pyppluss.fastqs import QuarticSolverVec as fastquartroots
from scipy.special import roots_legendre

#%% Base Models
def planet_hidden_area(radius_planet, x_star, y_star, star_planet_intersections, tol):
    """
    Calculating the area hidden by the planet.
    Models using DoubleHiddenAreaAlgorithm for consistency and lack of errors.
    
    Parameters
    ----------
    radius_planet : ``1-D array``
        Planet's radius relative to star's radius
    x_star : ``1-D array``
        x coordinate of the star's center
    y_star : ``1-D array``
        y coordinate of the star's center
    star_planet_intersections : ``2-D array``
        the array of star-planet intersections as returned by get_star_planet_intersection
    tol : ``float``, optional
        Error tolerance parameter
    Returns
    -------
    planet_area : ``1-D array``
        Area hidden by the planet
    """
    #The commented-out code is a case-specific implememtnation of Step 1.1.1 of the algorithm.
#    #Initilize with zeros to avoid special handling of another special case
#    planet_area = np.zeros_like(x_star)
#    calcr = radius_planet.copy()
#    calcr[calcr<1] = 1
#    #If there are <2 intersections and the planet is hiding part of the star
#    scalc = np.logical_and(np.any(np.isnan(star_planet_intersections),(1,2)),np.sqrt(x_star**2+y_star**2)<=calcr)
#    calcr = radius_planet[scalc].copy()
#    calcr[calcr>1] = 1
#    planet_area[scalc] = M.pi*calcr**2
#    #If there are 2 intersections, call the hidden area calculator.
#    scalc = np.any(np.isnan(star_planet_intersections),(1,2))
#    remcalc = np.logical_not(scalc)
#    border = border_sort(star_planet_intersections[remcalc])
#    pangles = border[:,3,:]
#    this_planet = Planet(radius_planet[remcalc], pangles, np.ones_like(pangles, dtype=bool))
#    this_star = Star(np.concatenate((x_star[remcalc].reshape((1,-1)),y_star[remcalc].reshape((1,-1)))), border[:,2,:], np.ones_like(pangles, dtype=bool))
#    planet_area[remcalc] = double_hidden_area((this_star,this_planet),border,tol)
    
    #Prepare the objects for the algorithm
    border = border_sort(star_planet_intersections)
    pangles = border[:,3,:]
    this_planet = Planet(radius_planet, pangles, np.ones_like(pangles, dtype=bool))
    this_star = Star(np.concatenate((x_star.reshape((1,-1)),y_star.reshape((1,-1)))), border[:,2,:], np.ones_like(pangles, dtype=bool))
    #Calculate the area
    planet_area = double_hidden_area((this_star,this_planet),border,tol)        
    return planet_area

def disk_hidden_area(disk_radius, disk_inclination, x_star, y_star, star_disk_intersections, tol):
    """
    Calculating the area hidden by the planet.
    Models using DoubleHiddenAreaAlgorithm for consistency and lack of errors.
    
    Parameters
    ----------
    disk_radius : ``1-D array``
        Disk's radius relative to star's radius
    disk_inclination : ``float``
        Disk's inclination (in radians)
    x_star : ``1-D array``
        x coordinate of the star's center
    y_star : ``1-D array``
        y coordinate of the star's center
    star_disk_intersections : ``2-D array``
        the array of star-disk intersections as returned by get_star_disk_intersection
    tol : ``float``, optional
        Error tolerance parameter
    Returns
    -------
    disk_area : ``1-D array``
        Area hidden by the disk
    """
    #The commented-out code is a case-specific implememtnation of Step 1.1.1 of the algorithm.
#    disk_area = np.zeros_like(x_star)
#    scalc = np.all(np.isnan(star_disk_intersections),(1,2))
#    simplei = scalc.copy()
#    sx = x_star[scalc]
#    sy = y_star[scalc]
#    sr = disk_radius[scalc]
#    sinc = M.sin(disk_inclination)
#    instar = sx**2+sy**2<=(1+tol)**2
#    indisk = np.sqrt((sx-sr*sinc)**2+sy**2)+np.sqrt((sx+sr*sinc)**2+sy**2)<=2*sr+tol
#    incalc = np.logical_or(instar,indisk)
#    sr = sr[incalc]**2*M.cos(disk_inclination)
#    sr[sr>1] = 1
#    scalc[np.nonzero(scalc)[0][np.logical_not(incalc)]] = False
#    disk_area[scalc] = M.pi*sr
#    scalc = simplei
#    remcalc = np.logical_not(scalc)
#    border = border_sort(star_disk_intersections[remcalc])
#    dangle = border[:,3,:]
#    this_disk = Disk(disk_radius[remcalc], disk_inclination, dangle, np.ones_like(dangle, dtype=bool))
#    this_star = Star(np.concatenate((x_star[remcalc].reshape((1,-1)),y_star[remcalc].reshape((1,-1)))), border[:,2,:], np.ones_like(dangle, dtype=bool))
#    disk_area[remcalc] = double_hidden_area((this_star,this_disk),border,tol)
    
    #Prepare the objects for the algorithm
    border = border_sort(star_disk_intersections)
    dangle = border[:,3,:]
    this_disk = Disk(disk_radius, disk_inclination, dangle, np.ones_like(dangle, dtype=bool))
    this_star = Star(np.concatenate((x_star.reshape((1,-1)),y_star.reshape((1,-1)))), border[:,2,:], np.ones_like(dangle, dtype=bool))
    #Calculate the area
    disk_area = double_hidden_area((this_star,this_disk),border,tol)
    
    return disk_area

def tot_hidden_area(radius_planet, radius_in, radius_out, x_star, y_star, ring_inclination, star_planet_intersections, star_disk_intersections_in, star_disk_intersections_out, disk_planet_intersections_in, disk_planet_intersections_out, opacity, tol=10**-10):
    """
    Calculation of the total hidden area of the planet and ring together
    
    Parameters
    ----------
    radius_planet : ``1-D array``
        Planet's radius
    radius_in : ``1-D array``
        Ring's inner radius
    radius_out : ``1-D array``
        Ring's outer radius
    x_star : ``1-D array``
        x coordinate of the star's center
    y_star : ``1-D array``
        y coordinate of the star's center
    ring_inclination : ``float``
        Ring's inclination (in radians)
    star_planet_intersections : ``2-D array``
        the array of star-planet intersections as returned by get_star_planet_intersection
    star_disk_intersections_in : ``2-D array``
        the array of star-inner disk intersections as returned by get_star_disk_intersection
    star_disk_intersections_out : ``2-D array``
        the array of star-outer disk intersections as returned by get_star_disk_intersection
    disk_planet_intersections_in : ``2-D array``
        the array of inner disk-planet intersections as returned by get_disk_planet_intersection
    disk_planet_intersections_out : ``2-D array``
        the array of outer disk-planet intersections as returned by get_disk_planet_intersection
    opacity : ``float``
        Ring's opacity
    tol : ``float``, optional
        Error tolerance parameter
    
    Returns
    -------
    hidden_area : ``1-D array``
        Total hidden area by the planet and ring.
    """
    #Planet hidden area
    planet_area = planet_hidden_area(radius_planet, x_star, y_star, star_planet_intersections, tol)
    #Disks hidden area
    disk_in_area = disk_hidden_area(radius_in, ring_inclination, x_star, y_star, star_disk_intersections_in, tol)
    disk_out_area = disk_hidden_area(radius_out, ring_inclination, x_star, y_star, star_disk_intersections_out, tol)
    #Double hidden area
    #Initial values assuming no intersections
    double_area_in = np.minimum(planet_area,disk_in_area)
    double_area_out = np.minimum(planet_area,disk_out_area)
    #When there are intersections, call the algorithm to find the double hidden area.
    calcin = np.logical_and(np.logical_and(planet_area>0,disk_in_area>0),np.any(np.logical_not(np.isnan(disk_planet_intersections_in)),(1,2)))
    star, planet, disk, dha_border_in = handler(radius_planet[calcin], radius_in[calcin], ring_inclination, x_star[calcin], y_star[calcin], star_planet_intersections[calcin], star_disk_intersections_in[calcin], disk_planet_intersections_in[calcin], tol)
    double_area_in[calcin] = double_hidden_area((star, planet, disk), dha_border_in, tol)
    calcout = np.logical_and(np.logical_and(planet_area>0,disk_out_area>0),np.any(np.logical_not(np.isnan(disk_planet_intersections_out)),(1,2)))
    star, planet, disk, dha_border_out = handler(radius_planet[calcout], radius_out[calcout], ring_inclination, x_star[calcout], y_star[calcout], star_planet_intersections[calcout], star_disk_intersections_out[calcout], disk_planet_intersections_out[calcout], tol)
    double_area_out[calcout] = double_hidden_area((star, planet, disk), dha_border_out, tol)
    #Conclusions
    ring_area = (disk_out_area-double_area_out)-(disk_in_area-double_area_in)
    hidden_area = opacity*ring_area+planet_area
    return hidden_area
#%% Paper Models
def uniform_rings(radius_planet, radius_in, radius_out, x_planet, y_planet, ring_inclination, ring_rotation, opacity=1.0, tol=10**-7):
    """LC value for planet with rings and uniform source
    
    Calculated using Polygon+Segments algorithm
    Assumes the input vectors are of the same length.

    Parameters
    ----------
    radius_planet : ``1-D array``
        Planet's radius
    radius_in : ``1-D array``
        Inner disk's radius
    radius_out : ``1-D array``
        Outer disk's radius
    x_planet : ``1-D array``
        x coordinate of the planet's center
    y_planet : ``1-D array``
        y coordinate of the planet's center
    ring_inclination : ``float``
        Inclination of the ring
    ring_rotation : ``float``
        Rotation of the ring
    opacity : ``float`` (0 to 1), optional
        Optical opacity for the ring hiding calculations.
    tol : ``float``, optional
        Error tolerance parameter
    Returns
    -------
    LC : ``1-D array``
        Estimated normalized light-curve value for the planet with rings.
    """
    #Input Validation
    LC = np.empty_like(x_planet)
    cond = np.isfinite(x_planet)
    LC[np.logical_not(cond)] = 1.0
    x_planet = x_planet[cond]
    y_planet = y_planet[cond]
    radius_in = radius_in[cond]
    radius_out = radius_out[cond]
    radius_planet = radius_planet[cond]
    #Swapping coordinate systems
    star_coords = coord_swap(x_planet, y_planet, ring_rotation)
    x_star = star_coords[0]
    y_star = star_coords[1]
    #Intersections
    star_planet_intersections = get_star_planet_intersection(x_star, y_star, radius_planet, tol)
    star_disk_intersections_in = get_star_disk_intersection(x_star, y_star, radius_in, ring_inclination, tol)
    star_disk_intersections_out = get_star_disk_intersection(x_star, y_star, radius_out, ring_inclination, tol)
    disk_planet_intersections_in = get_disk_planet_intersection(radius_planet, radius_in, ring_inclination, tol)
    disk_planet_intersections_out = get_disk_planet_intersection(radius_planet, radius_out, ring_inclination, tol)
    #Area Calculation & LC conclusions
    hidden_area = tot_hidden_area(radius_planet, radius_in, radius_out, x_star, y_star, ring_inclination, star_planet_intersections, star_disk_intersections_in, star_disk_intersections_out, disk_planet_intersections_in, disk_planet_intersections_out, opacity, tol)
    LC[cond] = 1-hidden_area/M.pi
    return LC


def LC_ringed(radius_planet, radius_in, radius_out, x_planet, y_planet, ring_inclination, ring_rotation, opacity, c1,c2,c3,c4, mucrit=0.0, n_center=5, n_gress=5, tol=10**-10):
    """LC value(s) for planet with rings and non uniform source
    
    Assuming quadratic limb darkening with paramters u1, u2
    Assign string value to c3 to signify quadratic LD
    
    Parameters
    ----------
    radius_planet : ``1-D array``
        Planet's radius
    radius_in : ``1-D array``
        Inner disk's radius
    radius_out : ``1-D array``
        Outer disk's radius
    x_planet : ``1-D array``
        x coordinate of the planet's center
    y_planet : ``1-D array``
        y coordinate of the planet's center
    ring_inclination : ``float``
        Inclination of the ring
    ring_rotation : ``float``
        Rotation of the ring
    opacity : ``float`` (0 to 1)
        Optical opacity for the ring hiding calculations.
    c1, c2 , c3, c4 : ``float``
        Limb darkening coefficients
    mucrit : ``float``, optional
        A critical mu, after which the LD is constant.
    n_cener,n_gress : ``int``, optional
        Order of Gaussian Quadrature in the integration
    tol : ``float``, optional
        Error tolerance parameter
    Returns
    -------
    LC : ``1-D array``
        Estimated normalized light-curve value for the planet with rings.
    """
    #Convert quadratic limb darkening coefficients into nonlinear
    if str(type(c3))=="<class 'str'>":
        c2 = c2+2*c4
        c4 = -c4
        c3 = 0
        c1 = 0        
    #Caclulate the edge of the interval of the integral
    rcrit = np.sqrt(1-mucrit**2)
    #Calculate Uniform-Source Value
    uniform_val = uniform_rings(radius_planet/rcrit, radius_in/rcrit, radius_out/rcrit, x_planet/rcrit, y_planet/rcrit, ring_inclination, ring_rotation, opacity, tol)
    #Integration
    #Function to be integrated is (c1+2*c2*t+3*c3*t**2+4*c4*t**3)*(uniform_rings(rp/rt,rin/rt,rout/rt,xp/rt,yp/rt,ring_inclination,ring_rotation, opacity, tol))*(1-t**4)
    #Where rt=sqrt(1-t**4)
    #Take stuff from uniform_val (e.g. intersection coords) and pass to integrate_swapped to save time
    y = integrate_swapped(n_center,n_gress, radius_planet, radius_in, radius_out, x_planet, y_planet, ring_inclination, ring_rotation, opacity, c1,c2,c3,c4, mucrit, tol, uniform_val)
    LC = (1-c1*(1-mucrit**0.5)-c2*(1-mucrit)-c3*(1-mucrit**1.5)-c4*(1-mucrit**2))*uniform_val+y
    #See integral in: https://www.wolframalpha.com/input/?i=integral+of+2x(1-c_1(1-(1-x%5E2)%5E(1%2F4))-c_2(1-(1-x%5E2)%5E(1%2F2))-c_3(1-(1-x%5E2)%5E(3%2F4))-c_4(1-(1-x%5E2)))
    star_area = -0.5*c4*rcrit**4-(c1+c2+c3-1)*rcrit**2-0.8*c1*(1-rcrit**2)**(5/4)-2/3*c2*(1-rcrit**2)**(3/2)-4/7*c3*(1-rcrit**2)**(7/4)+0.8*c1+2/3*c2+4/7*c3
    LC=LC/star_area
    return LC

def indefinite_integral_helper(x,c1,c2,c3,c4):
    """Indefinite Integral of for use in the following functions.
    The integral is after the swapping of t=(1-r^2)^(1/4). Assuming Fe=1.
    """
    return c1*x*(1-x**4/5)+c2*x**2*(1-x**4/3)+c3*x**3*(1-3/7*x**4)+c4*x**4*(1-0.5*x**4)

def integrated_swapped_end(t,c1,c2,c3,c4):
    """Integral from t to 1 of inetgral_swapped,
    assuming all area from uniform source is outside
    Remember t=1 is the center of the star and not the limb
    """
    return indefinite_integral_helper(1,c1,c2,c3,c4)-indefinite_integral_helper(t,c1,c2,c3,c4)
    
def integrated_swapped_start(t, c1, c2, c3, c4, c,a):
    """ Integral from a to t of inetgral_swapped,
    assuming all area from uniform source is already inside
    A_h=pi*(1-c) => Fe = 1-(1-c)/r^2
    => f(t) = I(r(t))*(1-(1-c)/(1-t^4))*(1-t^4) = I(r(t))*(1-t^4)-I(r(t))*(1-c)
    """
    return indefinite_integral_helper(t,c1,c2,c3,c4)-indefinite_integral_helper(a,c1,c2,c3,c4)-(1-c)*(c1*t+c2*t**2+c3*t**3+c4*t**4-c1*a-c2*a**2-c3*a**3-c4*a**4)

def vec_fixed_quad(x,w,n,a,b,rp,rin,rout,ring_inclination,ring_rotation,opacity,xp,yp,c1,c2,c3,c4,tol):
    """
    Calculating fixed-order Gaussian Quadrature integration sample values.
    """
    t = (np.dot(0.5*(b-a).reshape((-1,1)),(x+1).reshape((1,-1)))+np.repeat(a.reshape((-1,1)),n,1)).flatten(order='F')
    #Conversion to r
    rt = np.sqrt(1-t**4)
    #Tiling input vectors similarly to t.
    rp = np.repeat(rp.reshape((1,-1)),n,0).flatten()
    rin = np.repeat(rin.reshape((1,-1)),n,0).flatten()
    rout = np.repeat(rout.reshape((1,-1)),n,0).flatten()
    xp = np.repeat(xp.reshape((1,-1)),n,0).flatten()
    yp = np.repeat(yp.reshape((1,-1)),n,0).flatten()
    return ((c1+2*c2*t+3*c3*t**2+4*c4*t**3)*(uniform_rings(rp/rt,rin/rt,rout/rt,xp/rt,yp/rt,ring_inclination,ring_rotation, opacity, tol))*(1-t**4)).reshape((-1,n),order='F')

def integrate_swapped(n_center,n_gress,radius_planet, radius_in, radius_out, x_planet, y_planet, ring_inclination, ring_rotation, opacity, c1,c2,c3,c4, mucrit, tol,c):
    """Function for numerical integration: Calculating weights+function values and returning estimation of the integral.
    General case with changing uniform source.
    """
    rcrit = np.sqrt(1-mucrit**2)
    tcrit = np.sqrt(mucrit)
    #Find values reuired for  interval split calculations from helper functions
    star_coords = coord_swap(x_planet, y_planet, ring_rotation)
    x_star = star_coords[0]
    y_star = star_coords[1]
    mx = np.repeat(np.reshape(x_star,(len(x_star),1)),4,axis=1)
    my = np.repeat(np.reshape(y_star,(len(y_star),1)),4,axis=1)
    mrin = np.repeat(np.reshape(radius_in,(len(y_star),1)),4,axis=1)
    mrout = np.repeat(np.reshape(radius_out,(len(y_star),1)),4,axis=1)
    d = np.sqrt(x_planet**2+y_planet**2)
    #Calculate all of the potential interval splits
    splitr = np.empty((radius_planet.size,16))
    #Start and end of the interval
    splitr[:,0] = np.zeros_like(radius_planet)
    splitr[:,1] = np.ones_like(radius_planet)*rcrit
    #Minimum and maximum star-planet distance
    splitr[:,2] = abs(d-radius_planet)
    splitr[:,3] = d+radius_planet
    #Inner ring-planet intersections
    inner_intersections = get_disk_planet_intersection(radius_planet, radius_in, ring_inclination, tol)
    splitr[:,4:8] = np.sqrt((inner_intersections[:,0,:]-mx)**2+(inner_intersections[:,1,:]-my)**2)
    #Outer ring-planet intersections
    outer_intersections = get_disk_planet_intersection(radius_planet, radius_out, ring_inclination, tol)
    splitr[:,8:12] = np.sqrt((outer_intersections[:,0,:]-mx)**2+(outer_intersections[:,1,:]-my)**2)
 
    #Minimum and maximum inner ring-star distance
    #Solve the quartic
    coeff_a = radius_in*np.cos(ring_inclination)*y_star
    cond = coeff_a==0
    dists = np.empty((len(coeff_a),4),dtype=float)
    #Handler of special case
    #Here, we assume a ring exists:  0<radius_in*np.cos(ring_inclination) therefore cond is true iff y_star=0 iff the solutions for theta are 0 and pi.
    #Therefore the substituiton for t is ineffective, so we will use theta instead
    theta = np.empty((len(coeff_a[cond]),2),dtype=float)
    theta[:,0] = np.pi
    theta[:,1] = 0.0
    dists[cond,:2] = np.sqrt((my[cond,:2]-2*mrin[cond,:2]*np.cos(ring_inclination)*np.sin(theta))**2+(mx[cond,:2]-mrin[cond,:2]*np.cos(theta))**2)
    dists[cond,2:] = dists[cond,:2]

    #Quartic solver
    cond = np.logical_not(cond)
    coeff_b = 2*radius_in[cond]*x_star[cond]+2*radius_in[cond]**2*np.sin(ring_inclination)**2
    coeff_c = np.zeros_like(coeff_a[cond])
    coeff_d = 2*radius_in[cond]*x_star[cond]-2*radius_in[cond]**2*np.sin(ring_inclination)**2
    coeff_e = -coeff_a[cond]
    t = fastquartroots(coeff_a[cond].astype(complex),coeff_b.astype(complex),coeff_c.astype(complex),coeff_d.astype(complex),coeff_e.astype(complex)).T
    #Eliminate complex values
    t[np.iscomplex(t)] = np.nan
    t = t.real
    #Calculate distance at the extremum points
    dists[cond] = np.sqrt((my[cond]-2*mrin[cond]*np.cos(ring_inclination)*t/(1+t**2))**2+(mx[cond]-mrin[cond]*(1-t**2)/(1+t**2))**2)
    #Add the minimum and maximum to the interval split. (Ignoring NaNs)
    splitr[:,12] = np.nanmin(dists,-1)
    splitr[:,13] = np.nanmax(dists,-1)

    #Minimum and maximum outer ring-star distance
    #Solve the quartic: 2 a^2 t^3 - 2 a^2 t + 2 a c t^3 + 2 a c t - 2 b^2 t^3 + 2 b^2 t + b d t^4 - b d
#    + b d t^4 
#    +2 a^2 t^3 + 2 a c t^3 - 2 b^2 t^3 
#    
#    - 2 a^2 t + 2 a c t + 2 b^2 t 
#    - b d

    coeff_a = radius_out*np.cos(ring_inclination)*y_star
    cond = coeff_a==0
    dists = np.empty((len(coeff_a),4),dtype=float)
    #Handler of special case
    #Here, we assume a ring exists:  0<radius_in*np.cos(ring_inclination) therefore cond is true iff y_star=0 iff the solutions for theta are 0 and pi.
    #Therefore the substituiton for t is ineffective, so we will use theta instead
    theta = np.empty((len(coeff_a[cond]),2),dtype=float)
    theta[:,0] = np.pi
    theta[:,1] = 0.0
    dists[cond,:2] = np.sqrt((my[cond,:2]-2*mrout[cond,:2]*np.cos(ring_inclination)*np.sin(theta))**2+(mx[cond,:2]-mrout[cond,:2]*np.cos(theta))**2)
    dists[cond,2:] = dists[cond,:2]
    #Quartic solver
    cond = np.logical_not(cond)
    coeff_b = 2*radius_out[cond]*x_star[cond]+2*radius_out[cond]**2*np.sin(ring_inclination)**2
    coeff_c = np.zeros_like(coeff_a[cond])
    coeff_d = 2*radius_out[cond]*x_star[cond]-2*radius_out[cond]**2*np.sin(ring_inclination)**2
    coeff_e = -coeff_a[cond]
    t = fastquartroots(coeff_a[cond].astype(complex),coeff_b.astype(complex),coeff_c.astype(complex),coeff_d.astype(complex),coeff_e.astype(complex)).T
    #Eliminate complex values
    t[np.iscomplex(t)] = np.nan
    t = t.real
    #Calculate distance at the extremum points
    dists[cond] = np.sqrt((my[cond]-2*mrout[cond]*np.cos(ring_inclination)*t/(1+t**2))**2+(mx[cond]-mrout[cond]*(1-t**2)/(1+t**2))**2)
    #Add the minimum and maximum to the interval split. (Ignoring NaNs)
    splitr[:,14] = np.nanmin(dists,-1)
    splitr[:,15] = np.nanmax(dists,-1)

    #Ensure all splitr values are on the interval [0,rcrit]
    #Check whether there are points beyond one - whether there is a interval [a,rcrit] which requires numerical integration
    splitr_decider = np.zeros_like(splitr[:,2:]).astype(np.bool)
    splitr_decider[np.isnan(splitr[:,2:])] = True
    splitr_decider[~np.isnan(splitr[:,2:])] = splitr[:,2:][~np.isnan(splitr[:,2:])] < rcrit
    todo_start = np.all(splitr_decider, -1)
    #todo_start = np.all(np.logical_or(splitr[:,2:]<rcrit,np.isnan(splitr[:,2:])),-1)
    splitr[np.isnan(splitr)] = rcrit
    splitr[splitr>1] = rcrit
    splitr[splitr<0] = rcrit
    #by default sort acts on the last axis
    splitt = np.sort((1-splitr**2)**(1/4))
    #Calculating Legendre roots for n inputs
    xcen, wcen = roots_legendre(n_center)
    xgress, wgress = roots_legendre(n_gress)
    y = np.zeros_like(x_planet)
    #Calculate integral_swapped_start to initiate.
    #Use splitr instead of splitt in conditionals to avoid loss of accuracy
    start_ind = splitt.shape[1]-np.sum(splitr<rcrit,-1)
    y[todo_start] += integrated_swapped_start(splitt[todo_start,start_ind[todo_start]],c1,c2,c3,c4,c[todo_start],tcrit)
    #Test whether the star's center is NOT inside the planet or ring. Only then integral_swapped_start needs to be added.
    todo_end = np.logical_and(x_star**2+y_star**2>=radius_planet**2,np.logical_or(x_star**2+y_star**2/np.cos(ring_inclination)**2>=radius_out**2,x_star**2+y_star**2/np.cos(ring_inclination)**2<=radius_in**2))
    end_ind = np.sum(splitr>0.0,-1)-1
    y[todo_end] += integrated_swapped_end(splitt[todo_end,end_ind[todo_end]],c1,c2,c3,c4)
    for k in range(15):
        a = splitt[:,k]
        b = splitt[:,k+1]
        #Calculate the integral only when necessary: a,b are different and it is not solved via analytic method.
        i = np.logical_and(np.not_equal(a,b),np.logical_and(k>=start_ind-1+todo_start,k<=end_ind-todo_end))
        if np.any(i):
            icen = np.logical_and(i,todo_start)
            igress = np.logical_and(i,np.logical_not(todo_start))
            ficen = vec_fixed_quad(xcen,wcen,n_center,a[icen],b[icen],radius_planet[icen],radius_in[icen],radius_out[icen],ring_inclination,ring_rotation,opacity,x_planet[icen],y_planet[icen],c1,c2,c3,c4,tol)
            y[icen] += np.dot(ficen,wcen)*0.5*(b[icen]-a[icen])
            figress = vec_fixed_quad(xgress,wgress,n_gress,a[igress],b[igress],radius_planet[igress],radius_in[igress],radius_out[igress],ring_inclination,ring_rotation,opacity,x_planet[igress],y_planet[igress],c1,c2,c3,c4,tol)
            y[igress] += np.dot(figress,wgress)*0.5*(b[igress]-a[igress])
            
    return y



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun 18 10:57:19 2018

@author: edanrein

v14-May-2019
-Added to pyPplusS

v26-jun-2018
-New Code
"""

if __name__ == "__main__":
    from pyppluss.segment_models import LC_ringed as LC
    import numpy as np
    import matplotlib.pyplot as plt
    #number of points
    n=1000
    #Parameters for the test
    x_planet = np.linspace(-1.1,1.1,2*n)
    y_planet = np.zeros_like(x_planet)+0.0
    radius_planet = np.ones_like(x_planet)*0.11
    radius_in = np.ones_like(x_planet)*10000.44
    radius_out = np.ones_like(x_planet)*10000.7
    ring_inclination = 0
    ring_rotation = 0 
    opacity = 1.0
    #Range of orders for which the errors will be calculated
    orders = np.arange(3,15)
    #"True" order - this will be considered the true value
    trueorder = 20
    err = np.empty((len(orders),len(x_planet)))
    valstrue = vals = LC(radius_planet, radius_in, radius_out, x_planet, y_planet, ring_inclination, ring_rotation, opacity, 0.0, 0.448667, 0.0 ,0.313276,n_center=trueorder,n_gress=trueorder)
    plt.figure()
    for k in range(len(orders)):
        vals = LC(radius_planet, radius_in, radius_out, x_planet, y_planet, ring_inclination, ring_rotation, opacity, 0.0, 0.448667, 0.0 ,0.313276,n_center=orders[k],n_gress=orders[k])
        err[k] = abs(vals-valstrue)
        #Plotting
        plt.semilogy(x_planet,err[k],'o-')
        plt.text(x_planet[n],err[k][n],"n="+orders[k].__str__())
        
    plt.grid()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jul 25 16:04:06 2017
Object Oriented Double Hidden Area Algorithm
@author: edanrein
Implements methods from SectorAlgorithm to calculate double hidden area.

v10-Oct-2018
-Cleaned up

v02-Aug-2018
-Cleaned up

v26-May-2018
-Modified the point attribute to include two points
-Implemented use of two test points on the border of each curve instead of one
"""
from pyppluss.base_functions import get_coord_circle, get_coord_ellipse, get_arg, sector_area_disk, sector_area_planet, sector_area_star, get_dha_border, triangle_area
import math as M
import numpy as np
from pyppluss.base_functions import simpledist as edist

class Curve:
    """Parent class for all curves, with a few common methods.
    The Curves commonly implement a few methods:
        (a) Segment area calculation
        (b) Parameter storage
        (c) Parameter limitation (apply a logical index to the parameters)
    
    With regards to (c), the original parameter will be called i<param_name>.
    The limited parameter is called just <param_name>
        
    """    
    def coords(self, arg):
        raise NotImplementedError("overloaded by subclasses")
    
    def angle(self, k):
        """Looks up the angle of a point in input table, self.angles
        
        Parameters
        ----------
        k : ``int``
            Which point to examine
        """
        if np.isscalar(k):
            return self.angles[:,k]
        else:
            return self.angles[np.arange(self.angles.shape[0]),k]
    
    def sector_area(self, ang1, ang2):
        raise NotImplementedError("overloaded by subclasses")
        
    def segment_area(self, ang1, ang2):
        """Calculates segment area
        
        """
        #Calculate sector area using method from subcalss
        area_sec = self.sector_area(ang1,ang2)
        #Calculate triangle area
        p1 = self.coords(ang1)
        p2 = self.coords(ang2)
        area_tri = triangle_area(self.center[0], self.center[1], p1[0],p1[1], p2[0], p2[1])
        #Calculate segment, combining sector and triangle
        area = area_sec+area_tri*np.sign(ang2-ang1)*np.sign(abs(ang2-ang1)-M.pi)
        return area
            
    
    def gen_test(self, ang1, ang2):
        """Generate a test point, counterclockwise direction
        
        """
        test_arg = 0.5*(ang1+ang2)
        test_arg[ang1>ang2] += M.pi 
        return self.coords(test_arg)
    
    def is_on(self, ind):
        """Check using input self.pinds
        self.pinds has inedecies of all points from border on the curve.
        
        Parameters
        ----------
        ind: ``int``
            Which point to examine
        """
        if np.isscalar(ind):
            return self.pinds[:,ind]
        else:
            return self.pinds[np.arange(self.pinds.shape[0]),ind]
    
    def set_limiter(self,cond):
        raise NotImplementedError("overloaded by subclasses")
    
class Star(Curve):
    """
    
    """
    def __init__(self, center, inter_angles, on_inds):
        """Initialize with relevant parameters
        
        """
        self.icenter = center
        self.center = center
        self.iangles = inter_angles
        self.angles = inter_angles
        self.ipinds = on_inds
        self.pinds = on_inds
        self.ipoint = np.vstack((center,center)) + [[0],[1.],[0],[-1.]]
        self.point = self.ipoint
        self.iarea = np.pi*np.ones_like(center[0])
        self.area = self.iarea
    
    def coords(self, arg):
        """
        
        """
        return get_coord_circle(arg, self.center[0], self.center[1], 1)
    
    def sector_area(self, arg1, arg2):
        """
        
        """
        return sector_area_star(arg1,arg2)
    
    def is_in(self, point, tol=10**-8):        
        return edist(point, self.center)<=1+tol    

    
    def set_limiter(self,cond):
        """Required for vector handling. Allows conditionals.
        """
        self.center = self.icenter[:,cond]
        self.angles = self.iangles[cond]
        self.pinds = self.ipinds[cond]
        self.point = self.ipoint[:,cond]
        self.area = self.iarea[cond]

class Planet(Curve):
    """
    
    """
    def __init__(self, radius, inter_angles, on_inds):
        """Initialize this
        
        """
        self.radius = radius
        self.angles = inter_angles
        self.center = np.zeros((2,inter_angles.shape[0]))
        self.iradius = radius
        self.iangles = inter_angles
        self.icenter = self.center
        self.ipinds = on_inds
        self.pinds = on_inds
        self.ipoint = np.stack((np.zeros_like(radius),radius,np.zeros_like(radius),-radius),1).T
        self.point = self.ipoint
        self.iarea = np.pi*radius**2
        self.area=self.iarea
    
    def coords(self, arg):
        """
        
        """
        return get_coord_circle(arg, 0, 0, self.radius)
    
    def sector_area(self, arg1, arg2):
        """
        
        """
        return sector_area_planet(arg1, arg2, self.radius)
    
    def is_in(self, point, tol=10**-8):
        return edist(point, self.center)<=self.radius+tol
    
    def set_limiter(self,cond):
        """Required for vector handling. Allows conditionals.
        """
        self.center = self.icenter[:,cond]
        self.angles = self.iangles[cond]
        self.radius = self.iradius[cond]
        self.pinds = self.ipinds[cond]
        self.area = self.iarea[cond]
        self.point = self.ipoint[:,cond]
        
class Disk(Curve):
    """
    
    """
    def __init__(self, radius, inclination, inter_angles, on_inds):
        """Initialize this
        
        """
        self.radius = radius
        self.inclination = inclination
        self.angles = inter_angles
        self.center = np.zeros((2,inter_angles.shape[0]))
        self.iradius = radius
        self.iangles = inter_angles
        self.icenter = self.center
        self.foci = self.center.copy()
        self.foci[0] = self.radius*M.sin(self.inclination)
        self.ifoci = self.foci#.copy()
        self.ipinds = on_inds
        self.pinds = on_inds
        self.ipoint = np.stack((radius,np.zeros_like(radius),-radius,np.zeros_like(radius)),1).T
        self.point = self.ipoint
        self.iarea = np.pi*radius**2*M.cos(self.inclination)
        self.area=self.iarea

    def coords(self, arg):
        """
        
        """
        
        return get_coord_ellipse(arg, 0, 0, self.radius, self.radius*M.cos(self.inclination))
    
    def sector_area(self, arg1, arg2):
        """
        
        """
        return sector_area_disk(arg1, arg2, self.radius, self.inclination)
    
    def is_in(self, point, tol=10**-8):
        return edist(point, self.foci)+edist(point, -self.foci)<=2*self.radius+tol
    
    def set_limiter(self,cond):
        """Required for vector handling. Allows conditionals.
        """
        self.center = self.icenter[:,cond]
        self.angles = self.iangles[cond]
        self.radius = self.iradius[cond]
        self.foci = self.ifoci[:,cond]
        self.pinds = self.ipinds[cond]
        self.area = self.iarea[cond]
        self.point = self.ipoint[:,cond]

def intersection_area (curves,border, tol=10**-10):
    """Multiple hidden area calculations using object oriented, counterclockwise direction
    Assuming all border points are found before this.
    Assuming the same number of border points in all cases (since it is an array).
    In order to satisfy this, you can add NaNs to border and they will be ignored.
    
    Parameters
    ----------
    curves : ``list``
        A list of curve objects
    border : ``3-D array``
        Every split of the shape [k,:,:] is a single case, already oredered by angle relative to "center of mass".
    tol : ``float``
        Tolerance parameter
    
    Returns
    -------
    double_area : ``1-D array``
        Intersection area of curves.
    """
    seg_area = np.zeros(border.shape[0])
    poly_area = seg_area.copy()
    double_area = seg_area.copy()
    this_size = np.size(border,2)
    if not this_size in [2,3,4,5,6,7,8,9,10]:
        print(border)
        print(not list(border))
        print(this_size)
        raise Exception("Number of border points is not physically possible")
    maxn = border.shape[2]
    pnum = np.sum(np.logical_not(np.isnan(border[:,0,:])),-1)
    cases = border.shape[0]
    lastind = np.maximum(pnum-1,0)
    #Special case: When there are no intersections on the border of the double hidden area (Step 1.1.1)
    nodouble = lastind==0
    if np.any(nodouble):
        for c in curves:
            c.set_limiter(nodouble)
        for i in range(len(curves)):
            #When this curve's point is in all of the curves, set its area as the double-hidden area
            cond = np.empty((len(np.nonzero(nodouble)[0]),len(curves)),bool)
            for j in range(len(curves)):
                #Use Machine Precision as tolerance
                #Test only 1 point on the border. The use of machine precision reduces errors.
                #Unlike the DHA points, curves[i].point is known exactly (to machine precision) so there is no need for increased tolerance.
                #However, in certain cases there is a need for it since machine percision is relative error, i.e. if coordinates are ~200 instead of ~0.9 numerical errors occur.
                cond[:,j] = curves[j].is_in(curves[i].point[:2],tol=2.2204460492503131e-14)#16)
            condi = nodouble.copy()
            condi[nodouble] = np.all(cond,1)
            double_area[condi] = curves[i].area[np.all(cond,1)]
            #Otherwise, zero - double_area is initially set as zero and not changed.
    ydouble = np.logical_not(nodouble)
    poly_area = np.nansum(border[:,0,:-1]*border[:,1,1:]-border[:,1,:-1]*border[:,0,1:],-1)
    poly_area += border[np.arange(cases),0,lastind]*border[:,1,0]-border[np.arange(cases),1,lastind]*border[:,0,0]
    poly_area *= 0.5
    #Handling all-nans
    poly_area[np.isnan(poly_area)] = 0
    #The core loop of the algorithm
    #These two chunks of code are nearly identical, with one difference:
    #Since the top chunk only handles the 2-curve case, it needs one less conditional, checking whether an intersection point is on the curve.
    if len(curves)>2:
        for k in range(maxn):
            nind = k+np.ones(cases, dtype=int)
            nind[np.logical_or(k==maxn-1, k==lastind)] = 0
            is_done = np.isnan(border[:,0,k])
            #If there is no point to this, don't trigger the for loop at all
            if np.all(is_done):
                continue
            addarea = np.full((cases,len(curves)),np.inf)
            addarea[is_done] = 0
            for l in range(len(curves)):
                #Check whether the point and next_point are on curve
                #For increased efficiency, do not carry over all the vector, only a limited part.
                curve = curves[l]
                to_do = np.logical_not(is_done)
                curve.set_limiter(to_do)
                #We only consider cases where the curve connects the two points.
                to_curve = np.logical_and(curve.is_on(k), curve.is_on(nind[to_do]))
                if np.any(to_curve):
                    to_do[np.nonzero(to_do)[0][np.logical_not(to_curve)]] = False
                    curve.set_limiter(to_do)
                    ang = curve.angle(k)
                    next_ang = curve.angle(nind[to_do])
                    addarea[to_do,l] = curve.segment_area(ang, next_ang)
            seg_area += np.min(addarea,1)
    else:
        for k in range(maxn):
            nind = k+np.ones(cases, dtype=int)
            nind[np.logical_or(k==maxn-1, k==lastind)] = 0
            is_done = np.isnan(border[:,0,k])
            #If there is no point to this, don't trigger the for loop at all
            if np.all(is_done):
                continue
            addarea = np.full((cases,len(curves)),np.inf)
            addarea[is_done] = 0
            for l in range(len(curves)):
                #Check whether the point and next_point are on curve
                #For increased efficiency, do not carry over all the vector, only a limited part.
                curve = curves[l]
                to_do = np.logical_not(is_done)
                curve.set_limiter(to_do)
                #There are only 2 curves so all curves are on all points.
                ang = curve.angle(k)
                next_ang = curve.angle(nind[to_do])
                addarea[to_do,l] = curve.segment_area(ang, next_ang)
            seg_area += np.min(addarea,1)
    #Finalize the result.
    double_area[ydouble] = poly_area[ydouble]+seg_area[ydouble]
    return double_area

def border_sort(dha_border):
    """Sorting the border by angle relative to "center of mass"
    
    """
    maxn = dha_border.shape[2]
    pnum = np.sum(np.logical_not(np.isnan(dha_border[:,0,:])),-1)
    #If pnum=0, there are no points to begin with
    cond = np.logical_not(pnum==0)
    dborder = dha_border[cond]
    #Calculate the center of mass
    polyx = np.nansum(dborder[:,0,:], -1)/pnum[cond]
    polyy = np.nansum(dborder[:,1,:], -1)/pnum[cond]
    #Duplicate center of mass * (maximum number of points) for angle calculations
    xmat = np.repeat(polyx.reshape((-1,1)),maxn,1)
    ymat = np.repeat(polyy.reshape((-1,1)),maxn,1)
    #Sort the angles relative to the center of mass.
    angs = get_arg(dborder[:,0,:],dborder[:,1,:],xmat,ymat)
    inds = np.argsort(angs)
    dborder = np.moveaxis(dborder[np.arange(dborder.shape[0]),:,inds.T],0,2)
    dha_border[cond] = dborder
    return dha_border


def handler(radius_planet, disk_radius, disk_inclination, x_star, y_star, star_planet_intersections, star_disk_intersections, disk_planet_intersections, tol):
    """ Converts all the standard inputs into the appropriate classes (Star, Planet, Disk).
    """
    dha_border = get_dha_border(radius_planet, disk_radius, disk_inclination, x_star, y_star, star_planet_intersections, star_disk_intersections, disk_planet_intersections, tol)
    #Immediately re-order with respect to polygon center of mass.
    dha_border = border_sort(dha_border)
    indicators = dha_border[:,3:,:]
    is_star = (np.sum(indicators==0,1)>0)
    is_planet = (np.sum(indicators==1,1)>0)
    is_disk = (np.sum(indicators==2,1)>0)
    #Setup the curve instances
    sborder = np.swapaxes(dha_border,0,1).copy()
    pborder = sborder.copy()
    dborder = sborder.copy()
    sborder[:,np.logical_not(is_star)] = np.nan
    pborder[:,np.logical_not(is_planet)] = np.nan
    dborder[:,np.logical_not(is_disk)] = np.nan
    #Convert argument for star points to star argument
    sborder[2] = get_arg(sborder[0],sborder[1],np.repeat(x_star.reshape((-1,1)),10,1),np.repeat(y_star.reshape((-1,1)),10,1))
    star = Star(np.concatenate((x_star.reshape((1,-1)),y_star.reshape((1,-1)))), sborder[2], is_star)
    planet = Planet(radius_planet, pborder[2], is_planet)
    disk = Disk(disk_radius, disk_inclination, dborder[2], is_disk)
    
    return star, planet, disk, dha_border


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyppluss", # Replace with your own username
    version="0.1.5.2",
    author="Edan Rein",
    author_email="edanrein2000@gmail.com",
    description="Models for transiting exoplanets with rings",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/EdanRein/pyPplusS",
    license="MIT",
    packages=['pyppluss'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable"
    ],
    install_requires=[
        "numpy",
        "scipy"
    ]
)
